@@ -9,8 +9,8 @@ use solana_sdk::{
     account_utils::StateMut,
     bpf_loader_upgradeable::{self, UpgradeableLoaderState},
     feature_set::{
-        cpi_share_ro_and_exec_accounts, instructions_sysvar_enabled, updated_verify_policy,
-        FeatureSet,
+        cpi_share_ro_and_exec_accounts, demote_program_write_locks, instructions_sysvar_enabled,
+        updated_verify_policy, FeatureSet,
     },
     ic_msg,
     instruction::{CompiledInstruction, Instruction, InstructionError},
@@ -360,6 +360,8 @@ impl<'a> InvokeContext for ThisInvokeContext<'a> {
                 caller_write_privileges,
                 &mut self.timings,
                 self.feature_set.is_active(&updated_verify_policy::id()),
+                self.feature_set
+                    .is_active(&demote_program_write_locks::id()),
             ),
             None => Err(InstructionError::GenericError), // Should never happen
         }
@@ -572,6 +574,7 @@ impl MessageProcessor {
         instruction: &'a CompiledInstruction,
         executable_accounts: &'a [(Pubkey, Rc<RefCell<AccountSharedData>>)],
         accounts: &'a [Rc<RefCell<AccountSharedData>>],
+        demote_program_write_locks: bool,
     ) -> Vec<KeyedAccount<'a>> {
         let mut keyed_accounts = create_keyed_readonly_accounts(&executable_accounts);
         let mut keyed_accounts2: Vec<_> = instruction
@@ -582,7 +585,7 @@ impl MessageProcessor {
                 let index = index as usize;
                 let key = &message.account_keys[index];
                 let account = &accounts[index];
-                if message.is_writable(index) {
+                if message.is_writable(index, demote_program_write_locks) {
                     KeyedAccount::new(key, is_signer, account)
                 } else {
                     KeyedAccount::new_readonly(key, is_signer, account)
@@ -727,7 +730,14 @@ impl MessageProcessor {
     ) -> Result<(), InstructionError> {
         let invoke_context = RefCell::new(invoke_context);
 
-        let (message, executables, accounts, account_refs, caller_write_privileges) = {
+        let (
+            message,
+            executables,
+            accounts,
+            account_refs,
+            caller_write_privileges,
+            demote_program_write_locks,
+        ) = {
             let invoke_context = invoke_context.borrow();
 
             let caller_program_id = invoke_context.get_caller()?;
@@ -819,6 +829,7 @@ impl MessageProcessor {
                 accounts,
                 account_refs,
                 caller_write_privileges,
+                invoke_context.is_feature_active(&demote_program_write_locks::id()),
             )
         };
 
@@ -837,7 +848,7 @@ impl MessageProcessor {
             let invoke_context = invoke_context.borrow();
             for (i, (account, account_ref)) in accounts.iter().zip(account_refs).enumerate() {
                 let account = account.borrow();
-                if message.is_writable(i) && !account.executable {
+                if message.is_writable(i, demote_program_write_locks) && !account.executable {
                     account_ref.try_account_ref_mut()?.lamports = account.lamports;
                     account_ref.try_account_ref_mut()?.owner = account.owner;
                     if account_ref.data_len()? != account.data().len()
@@ -881,8 +892,15 @@ impl MessageProcessor {
                 Some(caller_write_privileges),
             )?;
             // Construct keyed accounts
-            let keyed_accounts =
-                Self::create_keyed_accounts(message, instruction, executable_accounts, accounts);
+            let demote_program_write_locks =
+                invoke_context.is_feature_active(&demote_program_write_locks::id());
+            let keyed_accounts = Self::create_keyed_accounts(
+                message,
+                instruction,
+                executable_accounts,
+                accounts,
+                demote_program_write_locks,
+            );
 
             // Invoke callee
             invoke_context.push(program_id)?;
@@ -954,6 +972,7 @@ impl MessageProcessor {
         rent: &Rent,
         timings: &mut ExecuteDetailsTimings,
         updated_verify_policy: bool,
+        demote_program_write_locks: bool,
     ) -> Result<(), InstructionError> {
         // Verify all executable accounts have zero outstanding refs
         Self::verify_account_references(executable_accounts)?;
@@ -972,7 +991,7 @@ impl MessageProcessor {
                 let account = accounts[account_index].borrow();
                 pre_accounts[unique_index].verify(
                     &program_id,
-                    message.is_writable(account_index),
+                    message.is_writable(account_index, demote_program_write_locks),
                     rent,
                     &account,
                     timings,
@@ -1004,6 +1023,7 @@ impl MessageProcessor {
         caller_write_privileges: Option<&[bool]>,
         timings: &mut ExecuteDetailsTimings,
         updated_verify_policy: bool,
+        demote_program_write_locks: bool,
     ) -> Result<(), InstructionError> {
         // Verify the per-account instruction results
         let (mut pre_sum, mut post_sum) = (0_u128, 0_u128);
@@ -1014,7 +1034,7 @@ impl MessageProcessor {
                 let is_writable = if let Some(caller_write_privileges) = caller_write_privileges {
                     caller_write_privileges[account_index]
                 } else {
-                    message.is_writable(account_index)
+                    message.is_writable(account_index, demote_program_write_locks)
                 };
                 // Find the matching PreAccount
                 for pre_account in pre_accounts.iter_mut() {
@@ -1110,8 +1130,15 @@ impl MessageProcessor {
             account_db,
             ancestors,
         );
-        let keyed_accounts =
-            Self::create_keyed_accounts(message, instruction, executable_accounts, accounts);
+        let demote_program_write_locks =
+            invoke_context.is_feature_active(&demote_program_write_locks::id());
+        let keyed_accounts = Self::create_keyed_accounts(
+            message,
+            instruction,
+            executable_accounts,
+            accounts,
+            demote_program_write_locks,
+        );
         self.process_instruction(
             program_id,
             &keyed_accounts,
@@ -1127,6 +1154,7 @@ impl MessageProcessor {
             &rent_collector.rent,
             timings,
             invoke_context.is_feature_active(&updated_verify_policy::id()),
+            demote_program_write_locks,
         )?;
 
         timings.accumulate(&invoke_context.timings);
@@ -2142,6 +2170,7 @@ mod tests {
         let programs: Vec<(_, ProcessInstructionWithContext)> =
             vec![(callee_program_id, mock_process_instruction)];
         let feature_set = FeatureSet::all_enabled();
+        let demote_program_write_locks = feature_set.is_active(&demote_program_write_locks::id());
 
         let ancestors = Ancestors::default();
         let mut invoke_context = ThisInvokeContext::new(
@@ -2180,7 +2209,7 @@ mod tests {
             .account_keys
             .iter()
             .enumerate()
-            .map(|(i, _)| message.is_writable(i))
+            .map(|(i, _)| message.is_writable(i, demote_program_write_locks))
             .collect::<Vec<bool>>();
         assert_eq!(
             MessageProcessor::process_cross_program_instruction(
@@ -2215,7 +2244,7 @@ mod tests {
                 .account_keys
                 .iter()
                 .enumerate()
-                .map(|(i, _)| message.is_writable(i))
+                .map(|(i, _)| message.is_writable(i, demote_program_write_locks))
                 .collect::<Vec<bool>>();
             assert_eq!(
                 MessageProcessor::process_cross_program_instruction(
